@@ -898,6 +898,9 @@ static int initiate_scan(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
+	if (task->path == NULL)
+		return -EINVAL;
+
 	if (task->scan_call != NULL)
 		return -EALREADY;
 
@@ -1508,6 +1511,8 @@ static int task_connect(struct supplicant_task *task)
 	unsigned int ssid_len;
 	int err;
 
+	connman_inet_ifup(task->ifindex);
+
 	address = connman_network_get_string(task->network, "Address");
 	security = connman_network_get_string(task->network, "WiFi.Security");
 	passphrase = connman_network_get_string(task->network, "WiFi.Passphrase");
@@ -1589,7 +1594,7 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		return;
 
 	if (task->scanning == TRUE && state != WPA_SCANNING) {
-		connman_device_set_scanning(task->device, FALSE);
+		connman_device_cleanup_scanning(task->device);
 		task->scanning = FALSE;
 	}
 
