@@ -1240,6 +1240,8 @@ int __connman_device_set_blocked(struct connman_device *device,
 	if (device->offlinemode == TRUE)
 		return 0;
 
+	connman_info("%s {rfkill} blocked %d", device->interface, blocked);
+
 	if (blocked == FALSE)
 		powered = device->powered_persistent;
 	else
@@ -1440,6 +1442,48 @@ void __connman_device_cleanup_networks(struct connman_device *device)
 					remove_unavailable_network, NULL);
 }
 
+static void scanning_changed(struct connman_device *device)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Scanning";
+
+	signal = dbus_message_new_signal(device->element.path,
+				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&device->scanning);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static void mark_network_available(gpointer key, gpointer value,
+						gpointer user_data)
+{
+	struct connman_network *network = value;
+
+	connman_network_set_available(network, TRUE);
+}
+
+void connman_device_cleanup_scanning(struct connman_device *device)
+{
+	device->scanning = FALSE;
+
+	scanning_changed(device);
+
+	g_hash_table_foreach(device->networks,
+				mark_network_available, NULL);
+}
+
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1450,10 +1494,6 @@ void __connman_device_cleanup_networks(struct connman_device *device)
 int connman_device_set_scanning(struct connman_device *device,
 						connman_bool_t scanning)
 {
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *key = "Scanning";
-
 	DBG("device %p scanning %d", device, scanning);
 
 	if (!device->driver || !device->driver->scan)
@@ -1464,21 +1504,7 @@ int connman_device_set_scanning(struct connman_device *device,
 
 	device->scanning = scanning;
 
-	signal = dbus_message_new_signal(device->element.path,
-				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &scanning);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
+	scanning_changed(device);
 
 	if (scanning == TRUE) {
 		reset_scan_trigger(device);
