@@ -89,6 +89,36 @@ char *connman_inet_ifname(int index)
 	return strdup(ifr.ifr_name);
 }
 
+short int connman_inet_ifflags(int index)
+{
+	struct ifreq ifr;
+	int sk, err;
+
+	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	if (sk < 0)
+		return -errno;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0) {
+		err = -errno;
+		goto done;
+	}
+
+	if (ioctl(sk, SIOCGIFFLAGS, &ifr) < 0) {
+		err = -errno;
+		goto done;
+	}
+
+	err = ifr.ifr_flags;
+
+done:
+	close(sk);
+
+	return err;
+}
+
 int connman_inet_ifup(int index)
 {
 	struct ifreq ifr;
@@ -288,6 +318,36 @@ static char *index2ident(int index, const char *prefix)
 	return str;
 }
 
+connman_bool_t connman_inet_is_mac80211(int index)
+{
+	connman_bool_t result = FALSE;
+	char phy80211_path[PATH_MAX];
+	struct stat st;
+	struct ifreq ifr;
+	int sk;
+
+	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	if (sk < 0)
+		return FALSE;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0)
+		goto done;
+
+	snprintf(phy80211_path, PATH_MAX,
+				"/sys/class/net/%s/phy80211", ifr.ifr_name);
+
+	if (stat(phy80211_path, &st) == 0 && (st.st_mode & S_IFDIR))
+		result = TRUE;
+
+done:
+	close(sk);
+
+	return result;
+}
+
 enum connman_device_type __connman_inet_get_device_type(int index)
 {
 	enum connman_device_type devtype = CONNMAN_DEVICE_TYPE_UNKNOWN;
@@ -309,11 +369,18 @@ enum connman_device_type __connman_inet_get_device_type(int index)
 	devname = ifr.ifr_name;
 
 	if (type == ARPHRD_ETHER) {
-		char bridge_path[PATH_MAX], wimax_path[PATH_MAX];
+		char phy80211_path[PATH_MAX];
+		char bonding_path[PATH_MAX];
+		char bridge_path[PATH_MAX];
+		char wimax_path[PATH_MAX];
 		struct stat st;
 		struct iwreq iwr;
 		char *devnode;
 
+		snprintf(phy80211_path, PATH_MAX,
+					"/sys/class/net/%s/phy80211", devname);
+		snprintf(bonding_path, PATH_MAX,
+					"/sys/class/net/%s/bonding", devname);
 		snprintf(bridge_path, PATH_MAX,
 					"/sys/class/net/%s/bridge", devname);
 		snprintf(wimax_path, PATH_MAX,
@@ -341,6 +408,10 @@ enum connman_device_type __connman_inet_get_device_type(int index)
 			devtype = CONNMAN_DEVICE_TYPE_WIMAX;
 		else if (stat(bridge_path, &st) == 0 && (st.st_mode & S_IFDIR))
 			devtype = CONNMAN_DEVICE_TYPE_UNKNOWN;
+		else if (stat(bonding_path, &st) == 0 && (st.st_mode & S_IFDIR))
+			devtype = CONNMAN_DEVICE_TYPE_UNKNOWN;
+		else if (stat(phy80211_path, &st) == 0 && (st.st_mode & S_IFDIR))
+			devtype = CONNMAN_DEVICE_TYPE_WIFI;
 		else if (ioctl(sk, SIOCGIWNAME, &iwr) == 0)
 			devtype = CONNMAN_DEVICE_TYPE_WIFI;
 		else
@@ -396,7 +467,7 @@ struct connman_device *connman_inet_create_device(int index)
 		name = strdup(devname);
 		break;
 	case CONNMAN_DEVICE_TYPE_MBM:
-		name = strdup(devname);
+		name = index2ident(index, "");
 		addr = index2addr(index);
 		node = __connman_udev_get_mbm_devnode(devname);
 		break;
@@ -455,8 +526,7 @@ done:
 	return device;
 }
 
-int connman_inet_set_address(int index, struct in_addr address,
-			struct in_addr netmask, struct in_addr broadcast)
+int connman_inet_set_address(int index, struct connman_ipaddress *ipaddress)
 {
 	struct ifreq ifr;
 	struct sockaddr_in addr;
@@ -478,7 +548,7 @@ int connman_inet_set_address(int index, struct in_addr address,
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
-	addr.sin_addr = address;
+	addr.sin_addr.s_addr = inet_addr(ipaddress->local);
 	memcpy(&ifr.ifr_addr, &addr, sizeof(ifr.ifr_addr));
 
 	err = ioctl(sk, SIOCSIFADDR, &ifr);
@@ -488,7 +558,7 @@ int connman_inet_set_address(int index, struct in_addr address,
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
-	addr.sin_addr = netmask;
+	addr.sin_addr.s_addr = htonl(~(0xfffffffflu >> ipaddress->prefixlen));
 	memcpy(&ifr.ifr_netmask, &addr, sizeof(ifr.ifr_netmask));
 
 	err = ioctl(sk, SIOCSIFNETMASK, &ifr);
@@ -498,7 +568,7 @@ int connman_inet_set_address(int index, struct in_addr address,
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
-	addr.sin_addr = broadcast;
+	addr.sin_addr.s_addr = inet_addr(ipaddress->broadcast);
 	memcpy(&ifr.ifr_broadaddr, &addr, sizeof(ifr.ifr_broadaddr));
 
 	err = ioctl(sk, SIOCSIFBRDADDR, &ifr);
