@@ -29,6 +29,9 @@
 
 #include "connman.h"
 
+#define MAX_CONNECT_RETRIES	2
+#define CONNECT_TIMEOUT		120
+
 static DBusConnection *connection = NULL;
 
 static GSequence *service_list = NULL;
@@ -48,11 +51,19 @@ struct connman_service {
 	connman_bool_t hidden;
 	connman_bool_t ignore;
 	connman_bool_t autoconnect;
+	connman_bool_t userconnect;
 	GTimeVal modified;
 	unsigned int order;
+	unsigned int failcounter;
 	char *name;
 	char *passphrase;
 	char *profile;
+	char *apn;
+	char *username;
+	char *password;
+	char *mcc;
+	char *mnc;
+	connman_bool_t roaming;
 	struct connman_ipconfig *ipconfig;
 	struct connman_device *device;
 	struct connman_network *network;
@@ -136,6 +147,12 @@ static const char *mode2string(enum connman_service_mode mode)
 		return "managed";
 	case CONNMAN_SERVICE_MODE_ADHOC:
 		return "adhoc";
+	case CONNMAN_SERVICE_MODE_GPRS:
+		return "gprs";
+	case CONNMAN_SERVICE_MODE_EDGE:
+		return "edge";
+	case CONNMAN_SERVICE_MODE_UMTS:
+		return "umts";
 	}
 
 	return NULL;
@@ -231,6 +248,36 @@ const char *__connman_service_default(void)
 	return type2string(service->type);
 }
 
+static void mode_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *str, *key = "Mode";
+
+	if (service->path == NULL)
+		return;
+
+	str = mode2string(service->mode);
+	if (str == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_STRING_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &str);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void state_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -291,6 +338,60 @@ static void strength_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void roaming_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Roaming";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->roaming);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static void autoconnect_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "AutoConnect";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->autoconnect);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void passphrase_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -342,15 +443,29 @@ static void passphrase_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
-static void autoconnect_changed(struct connman_service *service)
+static void apn_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
 	DBusMessageIter entry, value;
-	const char *key = "AutoConnect";
+	dbus_bool_t required;
+	const char *key = "SetupRequired";
 
 	if (service->path == NULL)
 		return;
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		return;
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		break;
+	}
+
+	required = (service->apn == NULL) ? TRUE : FALSE;
+
 	signal = dbus_message_new_signal(service->path,
 				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
 	if (signal == NULL)
@@ -362,8 +477,7 @@ static void autoconnect_changed(struct connman_service *service)
 
 	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
 					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
-							&service->autoconnect);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &required);
 	dbus_message_iter_close_container(&entry, &value);
 
 	g_dbus_send_message(connection, signal);
@@ -439,7 +553,38 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		connman_dbus_dict_append_variant(&dict, "Roaming",
+					DBUS_TYPE_BOOLEAN, &service->roaming);
+
+		if (service->mcc != NULL && service->mnc != NULL) {
+			connman_dbus_dict_append_variant(&dict, "MCC",
+					DBUS_TYPE_STRING, &service->mcc);
+			connman_dbus_dict_append_variant(&dict, "MNC",
+					DBUS_TYPE_STRING, &service->mnc);
+		}
+
+		if (service->apn != NULL) {
+			connman_dbus_dict_append_variant(&dict, "APN",
+					DBUS_TYPE_STRING, &service->apn);
+
+			if (service->username != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Username", DBUS_TYPE_STRING,
+							&service->username);
+
+			if (service->password != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Password", DBUS_TYPE_STRING,
+							&service->password);
+
+			required = FALSE;
+		} else
+			required = TRUE;
+
+		connman_dbus_dict_append_variant(&dict, "SetupRequired",
+						DBUS_TYPE_BOOLEAN, &required);
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		if (service->passphrase != NULL &&
@@ -499,7 +644,26 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	type = dbus_message_iter_get_arg_type(&value);
 
-	if (g_str_equal(name, "Passphrase") == TRUE) {
+	if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
+		connman_bool_t autoconnect;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->favorite == FALSE)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &autoconnect);
+
+		if (service->autoconnect == autoconnect)
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		service->autoconnect = autoconnect;
+
+		autoconnect_changed(service);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Passphrase") == TRUE) {
 		const char *passphrase;
 
 		if (type != DBUS_TYPE_STRING)
@@ -521,23 +685,63 @@ static DBusMessage *set_property(DBusConnection *conn,
 				"WiFi.Passphrase", service->passphrase);
 
 		__connman_storage_save_service(service);
-	} else if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
-		connman_bool_t autoconnect;
+	} else if (g_str_equal(name, "APN") == TRUE) {
+		const char *apn;
 
-		if (type != DBUS_TYPE_BOOLEAN)
+		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
-		if (service->favorite == FALSE)
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
 			return __connman_error_invalid_service(msg);
 
-		dbus_message_iter_get_basic(&value, &autoconnect);
+		dbus_message_iter_get_basic(&value, &apn);
 
-		if (service->autoconnect == autoconnect)
-			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+		g_free(service->apn);
+		service->apn = g_strdup(apn);
 
-		service->autoconnect = autoconnect;
+		apn_changed(service);
 
-		autoconnect_changed(service);
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Username") == TRUE) {
+		const char *username;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &username);
+
+		g_free(service->username);
+		service->username = g_strdup(username);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Password") == TRUE) {
+		const char *password;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &password);
+
+		g_free(service->password);
+		service->password = g_strdup(password);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
 
 		__connman_storage_save_service(service);
 	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
@@ -619,6 +823,9 @@ static connman_bool_t is_ignore(struct connman_service *service)
 	if (service->autoconnect == FALSE)
 		return TRUE;
 
+	if (service->roaming == TRUE)
+		return TRUE;
+
 	if (service->ignore == TRUE)
 		return TRUE;
 
@@ -661,8 +868,10 @@ void __connman_service_auto_connect(void)
 		iter = g_sequence_iter_next(iter);
 	}
 
-	if (service != NULL)
+	if (service != NULL) {
+		service->userconnect = FALSE;
 		__connman_service_connect(service);
+	}
 }
 
 static void reply_pending(struct connman_service *service, int error)
@@ -692,7 +901,7 @@ static void reply_pending(struct connman_service *service, int error)
 static gboolean connect_timeout(gpointer user_data)
 {
 	struct connman_service *service = user_data;
-	connman_bool_t auto_connect = FALSE;
+	connman_bool_t autoconnect = FALSE;
 
 	DBG("service %p", service);
 
@@ -715,12 +924,12 @@ static gboolean connect_timeout(gpointer user_data)
 		dbus_message_unref(service->pending);
 		service->pending = NULL;
 	} else
-		auto_connect = TRUE;
+		autoconnect = TRUE;
 
 	__connman_service_indicate_state(service,
 					CONNMAN_SERVICE_STATE_FAILURE);
 
-	if (auto_connect == TRUE)
+	if (autoconnect == TRUE && service->userconnect == FALSE)
 		__connman_service_auto_connect();
 
 	return FALSE;
@@ -752,6 +961,9 @@ static DBusMessage *connect_service(DBusConnection *conn,
 
 	service->ignore = FALSE;
 
+	service->userconnect = TRUE;
+	service->failcounter = 0;
+
 	service->pending = dbus_message_ref(msg);
 
 	err = __connman_service_connect(service);
@@ -786,6 +998,8 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 
 	DBG("service %p", service);
 
+	reply_pending(service, ECONNABORTED);
+
 	service->ignore = TRUE;
 
 	err = __connman_service_disconnect(service);
@@ -820,6 +1034,17 @@ static DBusMessage *remove_service(DBusConnection *conn,
 
 	passphrase_changed(service);
 
+	g_free(service->apn);
+	service->apn = NULL;
+
+	g_free(service->username);
+	service->username = NULL;
+
+	g_free(service->password);
+	service->password = NULL;
+
+	apn_changed(service);
+
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
@@ -941,6 +1166,11 @@ static void service_free(gpointer user_data)
 		service->ipconfig = NULL;
 	}
 
+	g_free(service->mcc);
+	g_free(service->mnc);
+	g_free(service->apn);
+	g_free(service->username);
+	g_free(service->password);
 	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
@@ -963,7 +1193,10 @@ void __connman_service_put(struct connman_service *service)
 
 		iter = g_hash_table_lookup(service_hash, service->identifier);
 		if (iter != NULL) {
+			reply_pending(service, ECONNABORTED);
+
 			__connman_service_disconnect(service);
+
 			g_sequence_remove(iter);
 		} else
 			service_free(service);
@@ -986,6 +1219,8 @@ static void __connman_service_initialize(struct connman_service *service)
 
 	service->ignore = FALSE;
 
+	service->userconnect = FALSE;
+
 	service->order = 0;
 }
 
@@ -1155,6 +1390,8 @@ int __connman_service_indicate_state(struct connman_service *service,
 		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
 		state_changed(service);
 
+		reply_pending(service, ECONNABORTED);
+
 		__connman_service_disconnect(service);
 	}
 
@@ -1166,6 +1403,9 @@ int __connman_service_indicate_state(struct connman_service *service,
 
 		reply_pending(service, 0);
 
+		service->userconnect = FALSE;
+		service->failcounter = 0;
+
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 
@@ -1179,8 +1419,19 @@ int __connman_service_indicate_state(struct connman_service *service,
 	}
 
 	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
+		if (service->failcounter++ < MAX_CONNECT_RETRIES) {
+			connman_warn("Connecting again (try %d)",
+						service->failcounter);
+			__connman_service_disconnect(service);
+			__connman_service_connect(service);
+			return 0;
+		}
+
 		reply_pending(service, EIO);
 
+		if (service->userconnect == FALSE)
+			__connman_service_auto_connect();
+
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 	} else
@@ -1244,8 +1495,16 @@ static connman_bool_t prepare_network(struct connman_service *service)
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+		break;
 	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
+		connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+		connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
 		break;
 	}
 
@@ -1270,7 +1529,10 @@ int __connman_service_connect(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn == NULL)
+			return -EINVAL;
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		switch (service->security) {
@@ -1310,7 +1572,7 @@ int __connman_service_connect(struct connman_service *service)
 			return err;
 		}
 
-		service->timeout = g_timeout_add_seconds(45,
+		service->timeout = g_timeout_add_seconds(CONNECT_TIMEOUT,
 						connect_timeout, service);
 
 		return -EINPROGRESS;
@@ -1325,8 +1587,6 @@ int __connman_service_disconnect(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	reply_pending(service, ECONNABORTED);
-
 	if (service->network != NULL) {
 		err = __connman_network_disconnect(service->network);
 	} else if (service->device != NULL) {
@@ -1530,6 +1790,8 @@ done:
 		service->passphrase = g_strdup(passphrase);
 	}
 
+	service->userconnect = TRUE;
+
 	err = __connman_service_connect(service);
 	if (err < 0 && err != -EINPROGRESS)
 		goto failed;
@@ -1882,6 +2144,24 @@ static enum connman_service_mode convert_wifi_security(const char *security)
 		return CONNMAN_SERVICE_SECURITY_UNKNOWN;
 }
 
+static enum connman_service_mode convert_cellular_mode(connman_uint8_t mode)
+{
+	switch (mode) {
+	case 0:
+	case 1:
+		return CONNMAN_SERVICE_MODE_GPRS;
+	case 3:
+		return CONNMAN_SERVICE_MODE_EDGE;
+	case 2:
+	case 4:
+	case 5:
+	case 6:
+		return CONNMAN_SERVICE_MODE_UMTS;
+	}
+
+	return CONNMAN_SERVICE_MODE_UNKNOWN;
+}
+
 static void update_from_network(struct connman_service *service,
 					struct connman_network *network)
 {
@@ -1907,6 +2187,7 @@ static void update_from_network(struct connman_service *service,
 	}
 
 	service->strength = connman_network_get_uint8(network, "Strength");
+	service->roaming = connman_network_get_bool(network, "Roaming");
 
 	str = connman_network_get_string(network, "WiFi.Mode");
 	service->mode = convert_wifi_mode(str);
@@ -1914,6 +2195,21 @@ static void update_from_network(struct connman_service *service,
 	str = connman_network_get_string(network, "WiFi.Security");
 	service->security = convert_wifi_security(str);
 
+	str = connman_network_get_string(network, "Cellular.MCC");
+	g_free(service->mcc);
+	service->mcc = g_strdup(str);
+
+	str = connman_network_get_string(network, "Cellular.MNC");
+	g_free(service->mnc);
+	service->mnc = g_strdup(str);
+
+	if (service->type == CONNMAN_SERVICE_TYPE_CELLULAR) {
+		connman_uint8_t value = connman_network_get_uint8(network,
+							"Cellular.Mode");
+
+		service->mode = convert_cellular_mode(value);
+	}
+
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
 		service->network = connman_network_ref(network);
@@ -1973,10 +2269,10 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		service->autoconnect = FALSE;
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		service->autoconnect = TRUE;
 		break;
 	}
@@ -2000,7 +2296,9 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 void __connman_service_update_from_network(struct connman_network *network)
 {
 	struct connman_service *service;
-	connman_uint8_t strength;
+	enum connman_service_mode mode;
+	connman_uint8_t strength, value;
+	connman_bool_t roaming;
 
 	service = __connman_service_lookup_from_network(network);
 	if (service == NULL)
@@ -2011,11 +2309,34 @@ void __connman_service_update_from_network(struct connman_network *network)
 
 	strength = connman_network_get_uint8(service->network, "Strength");
 	if (strength == service->strength)
-		return;
+		goto roaming;
 
 	service->strength = strength;
 
 	strength_changed(service);
+
+roaming:
+	roaming = connman_network_get_bool(service->network, "Roaming");
+	if (roaming == service->roaming)
+		goto done;
+
+	service->roaming = roaming;
+
+	roaming_changed(service);
+
+done:
+	if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+		return;
+
+	value = connman_network_get_uint8(service->network, "Cellular.Mode");
+	mode = convert_cellular_mode(value);
+
+	if (mode == service->mode)
+		return;
+
+	service->mode = mode;
+
+	mode_changed(service);
 }
 
 void __connman_service_remove_from_network(struct connman_network *network)
@@ -2124,6 +2445,15 @@ static int service_load(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->apn = g_key_file_get_string(keyfile,
+					service->identifier, "APN", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Username", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Password", NULL);
+
 		service->favorite = g_key_file_get_boolean(keyfile,
 				service->identifier, "Favorite", NULL);
 
@@ -2240,6 +2570,18 @@ update:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+							"APN", service->apn);
+
+		if (service->username != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Username", service->username);
+
+		if (service->password != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Password", service->password);
+
 		g_key_file_set_boolean(keyfile, service->identifier,
 					"Favorite", service->favorite);
 
